'
'  dd_dblib.bas
'
'  dblib wrapper functions for WinRep
'
'  >> 1-9-2002    DD
'     Added dblibInit/Done and error handling via callback.
'     Redid the way errors are handled.
'
'  >> 1-9-2002    DD
'     set max procs to be 1000 (%DBLIB_MAX_PROCS) in dblibInit
'
#if not %def(%DD_DBLIB_BAS)
%DD_DBLIB_BAS = 1

#include "win32api.inc"
#include "pb_sqldb.inc"
#include "dd_link.bas"
#include "dd_xml.bas"

%THRESHOLD_AUTOBIND_LENGTH = 1000
%MAX_AUTOBIND_LENGTH = 8000
%MIN_AUTOBIND_LENGTH = 50
%DBLIB_MAX_PROCS = 1000

%DBLIB_ERROR_SEVERITY = 1
%DBLIB_ERROR_DBCODE   = 2
%DBLIB_ERROR_OSCODE   = 3
%DBLIB_ERROR_DBSTRING = 4
%DBLIB_ERROR_OSSTRING = 5

'- The field types that I use ...
%DBLIB_FIELD_UID = 45
%DBLIB_FIELD_VARCHAR = 47
%DBLIB_FIELD_CHAR = 47
%DBLIB_FIELD_TEXT = 35
%DBLIB_FIELD_MONEY = 60
%DBLIB_FIELD_FLOAT = 62
%DBLIB_FIELD_REAL = 59
%DBLIB_FIELD_INT4 = 56
%DBLIB_FIELD_DATETIME = 61

Global g_dblib_CSInit as Long
Global g_dblib_CS as CRITICAL_SECTION
Global g_dblib_Error as Long

$DBLIB_DATABASE = "pappy:database-structure"
$DBLIB_TABLE = "pappy:table"
$DBLIB_FIELD = "pappy:column"
$DBLIB_INDEX = "pappy:index"

Type DBLIB_DATE_TYPE
   y as Long
   m as Long
   d as Long
   hr as Long
   mn as Long
   sc as Long
End Type

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblib_writelog
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
sub dblib_writelog ( ByVal sMsg as String )

#if 0
   Dim iFF as Long

   iFF = FreeFile
   open "c:\dbliblog.txt" for append as #iFF
   Print #iFF, Date$ + " " + Time$
   Print #iFF, sMsg
   close #iFF

#endif

end sub

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibField
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function dblibField ( ByVal pFields as Long, ByVal fieldName as String ) _
      as String

   Dim iCount as Long
   Dim iFound as Long
   Dim pNext as Long
   Dim sTypes as String
   Dim sFound as String

   fieldName = trim$(lcase$(fieldName))
   iCount = 0
   iFound = 0
   sTypes = llData(pFields)

   pNext = llNext(pFields)
   do until pNext = 0
      incr iCount
      if trim$(lcase$(llName(pNext))) = fieldName then
         iFound = iCount
         sFound = llData(pNext)
         exit do
      end if
      pNext = llNext(pNext)
   loop

   '- Didn't find the field we're looking for.
   if iFound < 1 then
      function = ""
      exit function
   end if

   Select Case asc(mid$(sTypes, iFound, 1))
      Case %DBLIB_FIELD_UID
         if len(trim$(sFound)) = 32 then
            '- Re-arrange the byte order
            sFound = mid$(sFound, 7, 2) + mid$(sFound, 5, 2) + _
                     mid$(sFound, 3, 2) + left$(sFound, 2) + "-" + _
                     mid$(sFound, 11, 2) + mid$(sFound, 9, 2) + "-" + _
                     mid$(sFound, 15, 2) + mid$(sFound, 13, 2) + "-" + _
                     mid$(sFound, 17, 2) + mid$(sFound, 19, 2) + "-" + _
                     right$(sFound, 12)
         end if

      Case %DBLIB_FIELD_MONEY
         Replace "," with "" in sFound
         Replace "$" with "" in sFound
      Case %DBLIB_FIELD_FLOAT
         Replace "," with "" in sFound
      Case %DBLIB_FIELD_REAL
         Replace "," with "" in sFound
      Case %DBLIB_FIELD_INT4
         Replace "," with "" in sFound
      Case %DBLIB_FIELD_DATETIME
         '- Don't do anything ... yet
      Case else
         '- Don't do anyting
   End Select
   Function = sFound
   'format$(iFound) + ":" + format$(asc(mid$(sTypes, iFound, 1))) + ":" + sFound

end function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibInitCritical
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Sub dblibInitCritical

   if g_dblib_CSInit = %false then
      g_dblib_CSInit = %true
      InitializeCriticalSection g_dblib_CS
   end if

End Sub

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibEnterCritial
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Sub dblibEnterCritical

   if g_dblib_CSInit = %false then
      dblibInitCritical
   end if

   if g_dblib_CSInit then
      EnterCriticalSection g_dblib_CS
   end if

End Sub

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibLeaveCritical
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Sub dblibLeaveCritical

   if g_dblib_CSInit then
      LeaveCriticalSection g_dblib_CS
   end if

End Sub

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibDeleteCritical
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Sub dblibDeleteCritical

   if g_dblib_CSInit then
      DeleteCriticalSection g_dblib_CS
      g_dblib_CSInit = %false
   end if

End Sub

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibHasError
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function dblibHasError Alias "dblibHasError" _
      ( ByVal dbProc as Long ) Export as Long

   Dim sErr as String

   dblibEnterCritical
   sErr = trim$(llDataByName(g_dblib_Error, format$(dbProc)))
   dblibLeaveCritical
   if sErr = "" then
      Function = %false
   else
      Function = %true
   end if

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibSetError
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Sub dblibSetError Alias "dblibSetError" _
      ( ByVal dbProc as Long, ByVal sMsg as String ) Export

   dblibEnterCritical
   llUpdateByName g_dblib_Error, Format$(dbProc), sMsg
   dblibLeaveCritical

End Sub

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibLastError
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function dblibLastError Alias "dblibLastError" _
      ( ByVal dbProc as Long ) Export as String

   dblibEnterCritical
   Function = llDataByName(g_dblib_Error, format$(dbProc))
   llUpdateByName g_dblib_Error, format$(dbProc), ""
   dblibLeaveCritical

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibDBError
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function dblibDBError Alias "dblibDBError" _
      ( ByVal dbProc as Long ) export as String

   Dim sErr as String

   sErr = dblibLastError(dbProc)
   function = parse$(sErr, $tab, %DBLIB_ERROR_DBCODE) + ":" + _
              parse$(sErr, $tab, %DBLIB_ERROR_DBSTRING)
End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibParseError
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function dblibParseError Alias "dblibParseError" _
      ( ByVal sErr as String, ByVal iWhich as Long ) export as String

   Function = parse$(sErr, $tab, iWhich)

End function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  internal_dblib_ErrorProc
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function internal_dblib_ErrorProc CDECL _
      ( ByVal dbProcess as Long, ByVal iSev as Long, ByVal dbErr as Long, _
        ByVal osErr as Long, zDBErrStr as Asciiz, zOSErrStr as Asciiz ) _
        as Long

   '- these are error attributed to logging in (dbopen)
   '  db-lib should be sending dbProcess as zero, but is not,
   '  so I change it to zero here.
   '
   if (dbErr = %SQLEMEM) or (dbErr = %SQLEPWD) or _
      (dbErr = %SQLECONN) or (dbErr = %SQLEDBPS) then
       dbProcess = 0
   end if

   dblibSetError dbProcess, _
         Format$(iSev) + $tab + _
         Format$(dbErr) + $tab + _
         Format$(osErr) + $tab + _
         zDBErrStr + $tab + _
         zOSErrStr
   function = %INT_CANCEL

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibInit
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Sub dblibInit Alias "dblibInit" Export

   dblibInitCritical
   dblibEnterCritical
      dbinit
      dbsetmaxprocs %DBLIB_MAX_PROCS
      dbsetopt %null, %DBTEXTLIMIT, "10000"    'format$(%MAX_AUTOBIND_LENGTH)
      dbsetopt %null, %DBTEXTSIZE, "10000"     'format$(%MAX_AUTOBIND_LENGTH)
      dberrhandle codeptr(internal_dblib_ErrorProc)
      g_dblib_Error = llAdd(0, "error_root", "error root")
   dblibLeaveCritical

End Sub

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibDone
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Sub dblibDone Alias "dblibDone" Export

   dblibEnterCritical
      if g_dblib_Error then llFree g_dblib_Error
      dbexit
   dblibLeaveCritical
   dblibDeleteCritical

End Sub

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibClean
'     Replaces single quotes (that's what i use) with their safe
'     equivalent.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function dblibClean Alias "dblibClean" _
      (incoming as String) Export as String

   Replace "'" with "''" in incoming
   Function = incoming

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibLogin
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function dblibLogin Alias "dblibLogin" _
      ( ByVal sUser as String, ByVal sPass as String ) Export as Long

   Dim pLogin as Long

   '- Init login structure
   pLogin = dblogin()
   if pLogin = 0 then
      Function = 0
      exit function
   end if

   dbsetlname pLogin, sUser + $NUL, %DBSETUSER
   dbsetlname pLogin, sPass + $NUL, %DBSETPWD
   dbsetlname pLogin, "dd Software", %DBSETAPP
   Function = pLogin

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibColInfo
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function dblibColInfo Alias "dblibColInfo" _
      ( ByVal dbProcess as Long, ByVal iColumn as Long, _
        rCol as DBCOL_TYPE ) Export as Long

   rCol.SizeOfStruct = len(rCol)
   'if dbcolinfo(dbProcess, %CI_REGULAR, iColumn, 0, rCol) = %SUCCEED then
   '   function = %true
   'else
   '   function = %false
   'end if
   dbcolinfo dbProcess, %CI_REGULAR, iColumn, 0, rCol
   function = %true

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibColName
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function dblibColName Alias "dblibColName" _
   ( ByVal dbProcess as Long, ByVal iCol as Long) Export as String

   Dim z as Asciiz Ptr
   z = dbcolnameZ(dbProcess, iCol)
   if z then
      Function = Trim$(lcase$(@z))
   else
      function = ""
   end if

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibColSource
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function dblibColSource Alias "dblibColSource" _
   ( ByVal dbProcess as Long, ByVal iCol as Long) Export as String

   Dim z as Asciiz Ptr

   z = dbcolsourceZ(dbProcess, iCol)
   if z then
      function = trim$(lcase$(@z))
   else
      function = ""
   end if

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibColTable
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function dblibColTable Alias "dblibColTable" _
      ( ByVal dbProcess as Long, ByVal iCol as Long ) Export as String

   Dim iReturn as Long
   Dim z as Asciiz Ptr

   z = dbtabsourceZ(dbProcess, iCol, iReturn)
   if z then
      function = trim$(lcase$(@z))
   else
      function = ""
   end if

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibColsToList
'
'  Gets all of the column titles for the current result
'  set and puts them in a linked list. Returns a pointer
'  to the linked list.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function dblibColsToList Alias "dblibColsToList" _
      ( ByVal dbProcess as Long ) Export as DWord

   Dim i as Long
   Dim iCols as Long
   Dim pCol as LINKED_LIST_TYPE PTR

   iCols = dbnumcols(dbProcess)
   for i = 1 to iCols
      if pCol = 0 then
         pCol = llAdd(pCol, Format$(i), dblibcolname(dbProcess, i))
      else
         llAdd pCol, Format$(i), dblibcolname(dbProcess, i)
      end if
   next i

   Function = pCol

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibAutobind
'
'  Returns a pointer to the data row linked list. Be sure to
'  pass it the pointer to the column headers (retrieved by calling
'  dblibColsToList).
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function dblibAutobind Alias "dblibAutobind" _
      ( ByVal dbProcess as Long ) Export as Dword

   Dim i as Long
   Dim iCols as Integer
   Dim iLen as Long
   Dim iType as Long
'   Dim sBind as String
   Dim sField as String
   Dim sTypes as String
   Dim pData as LINKED_LIST_TYPE PTR
   Dim pThis as LINKED_LIST_TYPE PTR

   pData = llAdd(0, "*types*", "")
   pThis = pData

   '- Loop through all of the files
   iCols = dbnumcols(dbProcess)
   for i = 1 to iCols
      sField = dblibcolname(dbProcess, i)

      '- Pappy field kludge
      if sField = "agen_repcomm_use_saletype_rate" then
         sField = "agen_repcomm_use_saletype_rate_yn"
      end if
      iLen = dbcollen(dbProcess, i) + 1
      iType = dbcoltype(dbProcess, i)

      sTypes = sTypes + chr$(iType)
      if iLen > %THRESHOLD_AUTOBIND_LENGTH then
         iLen = %MAX_AUTOBIND_LENGTH
      elseif iLen < %MIN_AUTOBIND_LENGTH then
         iLen = %MIN_AUTOBIND_LENGTH
      elseif iLen < 1 then
         iLen = 1
      end if

      'stdout sField + " requires " + format$(iLen) + " bytes."
      'sBind = space$(iLen)
      'dblib_writelog "field=" + sField
      pThis = llAdd(pThis, sField, space$(iLen))
      'stdout "len(llData(pThis)) = " + format$(len(llData(pThis)))

      '- Bind the field
      if pThis then
         if @pThis.pData then
            dbbind dbProcess, i, %NTBSTRINGBIND, iLen, ByVal @pThis.pData
         end if
      end if
   next i

   llUpdate pData, sTypes

   Function = pData

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibUseDB
'
'  Returns dbProc (0 on failure).
'  Uses the database pointed to by sDatabase
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function dblibUseDB(ByVal sServer as String, ByVal sDatabase as String, _
                    ByVal sUser as String, ByVal sPassword as String ) as Long

   Dim dbProc as Long
   Dim pLogin as Long

   dblib_writelog "Trying to login: " + $crlf + _
                  "     server=" + sServer + $crlf + _
                  "  sDatabase=" + sDatabase + $crlf + _
                  "      sUser=" + sUser + $crlf + _
                  "  sPassword=" + sPassword

   '- Create the login structure
   pLogin = dblibLogin(sUser, sPassword)
   if pLogin = 0 then
      Function = 0
      exit function
   end if

   dblib_writelog "dbopen is next"

   '- Connect to the server
   dbProc = dbopen(pLogin, sServer + $NUL)
   if dbProc then


      dblib_writelog "dbuse is next"

      '- If we don't want to use a database, then
      '  just return.
      '
      if sDatabase = "" then
         'dbsetopt dbProc, %DBTEXTLIMIT, "10000"    'format$(%MAX_AUTOBIND_LENGTH)
         'dbsetopt dbProc, %DBTEXTSIZE, "10000"     'format$(%MAX_AUTOBIND_LENGTH)
         Function = dbProc

      '- Try to select the requested database.
      '  If this fails, the function fails.
      '
      else
         if dbuse(dbProc, sDatabase + $NUL) = 0 then
            dbclose dbProc
            dblib_writelog "dbuse is zero"
            Function = 0
         else
            'dbsetopt dbProc, %DBTEXTLIMIT, "10000"    'format$(%MAX_AUTOBIND_LENGTH)
            'dbsetopt dbProc, %DBTEXTSIZE, "10000"     'format$(%MAX_AUTOBIND_LENGTH)
            Function = dbProc
         end if
      end if
   else
      dblib_writelog "dbproc as zero"
      Function = 0
   end if

   '- Cleanup the login structure
   if pLogin then dbFreeLogin pLogin

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibRestoreBackup
'
'  >  Restores a database from a backup file.
'  >  Returns %true if successful, and %false if not.
'
'  >  Parameters
'        originalDB = name of the original db file (e.g. pappy_data)
'        originalLog = name of the original log file (e.g. pappy_log)
'        newDB = name of the new db (copy_of_pappy)
'        backupFile = full dos path to backup (c:\backup_pappy.bak)
'        newDir = directory for the new data file to be stored.
'
'        sError = returns with an error message
'        sSQL = returns with the SQL exectuted
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function dblibRestoreBackup Alias "dblibRestoreBackup" _
      ( ByVal dbProc as Long, _
        ByVal originalDB as String, ByVal originalLog as String, _
        ByVal newDB as String, ByVal backupFile as String, _
        ByVal newDir as String, sError as String, _
        sSQL as String) Export as Long

   Dim iCount as Long
   Dim iReturn as Long

   sError = ""

   originalDB = trim$(lcase$(originalDB), ".mdf")
   originalLog = trim$(lcase$(originalLog), ".ldf")

   '- Destination directory
   newDir = trim$(newDir)
   if right$(newDir, 1) <> "\" then newDir = newDir + "\"

   sSQL = "RESTORE DATABASE " + newDB + $crlf + _
          " FROM DISK = '" + backupFile + "'" + $crlf + _
          " WITH MOVE '" + originalLog + "'" + $crlf + _
          " TO '" + newDir + newDB + ".ldf', " + $crlf + _
          " MOVE '" + originalDB + "'" + $crlf + _
          " TO '" + newDir + newDB + ".mdf', RECOVERY" + $crlf

   dbcmd dbProc, sSQL + $nul

   '- Execute the sql statement
   if dbsqlexec(dbProc) = %FAIL then
      sError = "dblibRestoreBackup Failed: dbsqlexec"
      function = %false
      exit function
   end if

   iReturn = dbresults(dbProc)
   if iReturn <> %SUCCEED then
      dbcanquery dbProc
      function = %false
      exit function
   end if

   '- Wait for more results
   do until iReturn = %NO_MORE_RESULTS
      apiSleep 5
      iReturn = dbresults(dbProc)
   loop

   Function = %true

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibExecuteOne
'
'  >  Executes an sql query.
'  >  Returns a pointer (linked list) to the bound columns.
'     or Returns 0 on failure.
'  >  This is designed to only return one row, the programmer
'     doesn't have to call dbcanquery or dbnextrow as this routine
'     will clean up after itself.
'  >  The programmer must call llFree on the pointer to the
'     autobound columns.
'  >  Fills iFound with %true if a column was found. Fills it with
'     %false if not.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function dblibExecuteOne Alias "dblibExecuteOne" _
      (  ByVal dbProc as Long, ByVal sDB as String, _
         ByVal sSQL as String, iFound as Long) as Dword

   Dim iReturn as Long
   Dim pCol as LINKED_LIST_TYPE PTR
   Dim pData as LINKED_LIST_TYPE PTR

   iFound = %false

   if sDB <> "" then
      if dbuse(dbProc, sDB + $nul) = %FAIL then
         function = 0
         exit function
      end if
   end if

   '- Execute the sql
   dbcmd dbProc, sSQL + $nul

   '- Execute the sql statement
   if dbsqlexec(dbProc) = %FAIL then
      function = 0
      exit function
   end if

   '- Prep the results
   iReturn = dbresults(dbProc)
   if iReturn <> %SUCCEED then
      dbcanquery dbProc
      function = %false
      exit function
   end if

   '- Bind the table
   pData = dblibAutoBind(dbProc)

   '- Get the first row
   iReturn = dbnextrow(dbProc)

   '- Did we find any rows?
   if iReturn <> %NO_MORE_ROWS then
      dbcanquery dbProc
      iFound = %true
   end if
   function = pData

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibExecute
'
'  >  Executes an sql query.
'  >  Returns a pointer (linked list) to the bound columns.
'     or Returns 0 on failure.
'  >  The programmer is responsible for calling dbnextrow
'     until it returns %NO_MORE_ROWS. This call (unlike dblibExecuteOne)
'     leaves the sql statement open.
'  >  The programmer must call llFree on the pointer to the
'     autobound columns.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function dblibExecute Alias "dblibExecute" _
      (  ByVal dbProc as Long, ByVal sDB as String, _
         ByVal sSQL as String) as Dword

   Dim iReturn as Long

   if trim$(sDB) <> "" then
      if dbuse(dbProc, sDB + $nul) = %FAIL then
         function = 0
         exit function
      end if
   end if

   '- Execute the sql
   dbcmd dbProc, sSQL + $nul

   '- Execute the sql statement
   if dbsqlexec(dbProc) = %FAIL then
      function = 0
      exit function
   end if

   '- Prep the results
   iReturn = dbresults(dbProc)
   if (iReturn <> %SUCCEED) then
      dbcanquery dbProc
      function = %false
      exit function
   end if

'- ADDED, but maybe remove
'   do until iReturn = %NO_MORE_RESULTS
'      apiSleep 5
'      iReturn = dbresults(dbProc)
'   loop

   '- Bind the table
   function = dblibAutoBind(dbProc)

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibExecuteNoReturn
'
'  >  Executes an sql query that expects no result set.
'     E.G. INSERT/UPDATE/DELETE queries
'  >  Returns %true if successful, %false if not
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function dblibExecuteNoReturn Alias "dblibExecuteNoReturn" _
      (  ByVal dbProc as Long, ByVal sDB as String, _
         ByVal sSQL as String) as Long

   Dim iReturn as Long

   if trim$(sDB) <> "" then
      if dbuse(dbProc, sDB + $nul) = %FAIL then
         function = 0
         exit function
      end if
   end if

   '- Execute the sql
   dbcmd dbProc, sSQL + $nul

   '- Execute the sql statement
   if dbsqlexec(dbProc) = %FAIL then
      function = 0
      exit function
   end if

   '- Prep the results
   iReturn = dbresults(dbProc)
   if iReturn <> %SUCCEED then
      dbcanquery dbProc
      function = %false
      exit function
   end if

   '- Get the first row
   iReturn = dbnextrow(dbProc)
   do until iReturn = %NO_MORE_ROWS
      apiSleep 1
   loop
   function = %true

End Function


'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibTableExists
'
'  Returns %true if the table specified exists in the database specified.
'  Otherwise it returns %false. If returning %false, then last error should
'  be blank. If not, then the table might exists, but there was an error.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function dblibTableExists Alias "dblibTableExists" _
      ( ByVal dbProc as Long, ByVal sDB as String, ByVal sTable as String) _
        Export as Long

   Dim iFound as Long
   Dim pData as Dword
   Dim sSQL as String

   if dbuse(dbProc, sDB + $nul) = %FAIL then
      function = %false
      exit function
   end if

   sSQL = "SELECT TABLE_NAME FROM INFORMATION_SCHEMA.TABLES WHERE " + _
          "TABLE_NAME='" + sTable + "'"

   pData = dblibExecuteOne(dbProc, sDB, sSQL, iFound)
   if pData = 0 then
      function = %false
   elseif iFound = %false then
      function = %false
   else
      llFree pData
      function = %true
   end if

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibEnumDatabases
'
'  Returns a pointer to a list of database names and their sizes.
'  The name of the database is the name of the list item.
'  The size of the database is the data of the list item.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function dblibEnumDatabases Alias "dblibEnumDatabases" _
      ( ByVal dbProc as Long ) Export as DWord

   Dim i as Long
   Dim iCount as Long
   Dim iReturn as Long
   Dim pDB as Long
   Dim zDBName as Asciiz * 1000
   Dim zDBSize as Asciiz * 100

   '- The name of the stored proc.
   if dbrpcinit(dbProc, "sp_databases", 0) = %false then
      function = 0
      exit function
   end if

   '- Execute it
   if dbrpcexec(dbProc) = %FAIL then
      function = 0
      exit function
   end if

   '- Process the results
   iReturn = dbresults(dbProc)
   if iReturn <> %SUCCEED then
      dbcanquery dbProc
      function = %false
      exit function
   end if

   '- Bind the column that we care about
   For i = 1 to dbnumcols(dbProc)
      Select case lcase$(trim$(dblibColName(dbProc, i)))
         Case "database_name"
            dbbind dbProc, i, %NTBSTRINGBIND, 0, zDBName
         Case "database_size"
            dbbind dbProc, i, %NTBSTRINGBIND, 0, zDBSize
      End Select
   Next i

   iCount = 0
   iReturn = dbnextrow(dbProc)
   Do until (iReturn = %NO_MORE_ROWS) or (iReturn = %FAIL)
      incr iCount
      if pDB = 0 then
         pDB = llAdd(0, trim$(zDBName), trim$(zDBSize))
      else
         llUpdateByName pDB, trim$(zDBName), trim$(zDBSize)
      end if
      iReturn = dbnextrow(dbProc)
   Loop
   Function = pDb

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibEnumTables
'
'  Returns a list of database table names.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function dblibEnumTables Alias "dblibEnumTables" _
      ( ByVal dbProc as Long, ByVal dbName as String ) Export as Dword

   Dim i as Long
   Dim iBind as Long
   Dim iCount as Long
   Dim iReturn as Long
   Dim zTableName as Asciiz * 1000
   Dim zType as Asciiz * 1000
   Dim pTable as LINKED_LIST_TYPE PTR

   '- Select the database
   if dbuse(dbProc, dbName + $nul) = %FAIL then
      function = 0
      exit function
   end if

   '- The name of the stored proc.
   if dbrpcinit(dbProc, "sp_tables", 0) = %false then
      function = 0
      exit function
   end if

   '- Execute it
   if dbrpcexec(dbProc) = %FAIL then
      function = 0
      exit function
   end if

   '- Process the results
   iReturn = dbresults(dbProc)
   if iReturn <> %SUCCEED then
      dbcanquery dbProc
      function = %false
      exit function
   end if

   '- Bind the column that we care about
   For i = 1 to dbnumcols(dbProc)
      Select case lcase$(trim$(dblibColName(dbProc, i)))
         Case "table_name"
            dbbind dbProc, i, %NTBSTRINGBIND, 0, zTableName
         Case "table_type"
            dbbind dbProc, i, %NTBSTRINGBIND, 0, zType
      End Select
   Next i

   iCount = 0
   iReturn = dbnextrow(dbProc)
   Do until iReturn = %NO_MORE_ROWS
      if (lcase$(trim$(zType)) = "table") and _
         (lcase$(trim$(zTableName)) <> "dtproperties") then
         incr iCount
         if pTable = 0 then
            pTable = llAdd(0, Format$(iCount), lcase$(trim$(zTableName)))
         else
            llUpdateByName pTable, Format$(iCount), lcase$(trim$(zTableName))
         end if
      end if
      iReturn = dbnextrow(dbProc)
   Loop

   Function = pTable

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibEnumFields
'
'  Fields in the list are formatted like:
'  value="fieldname;type;precision;length"
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function dblibEnumFields Alias "dblibEnumFields" _
      ( ByVal dbProc as Long, ByVal dbName as String, _
        ByVal tableName as String ) Export as DWord

   Dim i as Long
   Dim iCount as Long
   Dim iReturn as Long
   Dim sAccum as String
   Dim zField as Asciiz * 300
   Dim zType as Asciiz * 300
   Dim zPrec as Asciiz * 300
   Dim zLen as Asciiz * 300
   Dim zScale as Asciiz * 300
   Dim zNulls as Asciiz * 300
   Dim zParam as Asciiz * 300
   Dim pField as LINKED_LIST_TYPE PTR

   '- Select the database
   if dbuse(dbProc, dbName + $nul) = %FAIL then
      function = 0
      exit function
   end if

   '- The name of the stored proc.
   if dbrpcinit(dbProc, "sp_columns", 0) = %false then
      function = 0
      exit function
   end if

   '- Specify the parameters
   zParam = trim$(tableName)
   if dbrpcparam(dbProc, "@table_name", 0, %SQLVARCHAR, _
         sizeof(zParam), len(zparam), ByVal VarPtr(zParam)) = %FAIL then
      function = 0
      exit function
   end if

   '- Execute it
   if dbrpcexec(dbProc) = %FAIL then
      function = 0
      exit function
   end if

   '- Process the results
   iReturn = dbresults(dbProc)
   if iReturn <> %SUCCEED then
      dbcanquery dbProc
      function = %false
      exit function
   end if

   '- Bind the column that we care about
   For i = 1 to dbnumcols(dbProc)
      select case trim$(lcase$(dblibColname(dbProc, i)))
         Case "column_name"
            dbbind dbProc, i, %NTBSTRINGBIND, 0, zField
         Case "type_name"
            dbbind dbProc, i, %NTBSTRINGBIND, 0, zType
         Case "length"
            dbbind dbProc, i, %NTBSTRINGBIND, 0, zlen
         Case "precision"
            dbbind dbProc, i, %NTBSTRINGBIND, 0, zPrec
         Case "scale"
            dbbind dbProc, i, %NTBSTRINGBIND, 0, zScale
         Case "nullable"
            dbbind dbProc, i, %NTBSTRINGBIND, 0, zNulls
      end select
   Next i

   iCount = 0
   iReturn = dbnextrow(dbProc)
   Do until iReturn = %NO_MORE_ROWS
      incr iCount

      sAccum = lcase$(trim$(zField)) + ";" + lcase$(trim$(zType)) + ";" + _
               trim$(zLen) + ";" + trim$(zPrec) + ";" + trim$(zScale) + ";" + _
               trim$(zNulls)

      if pField = 0 then
         pField = llAdd(0, trim$(zField), sAccum)
      else
         llUpdateByName pField, trim$(zField), sAccum
      end if
      iReturn = dbnextrow(dbProc)
   Loop

   Function = pField

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibEnumIndices
'
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function dblibEnumIndices Alias "dblibEnumIndices" _
      ( ByVal dbProc as Long, ByVal dbName as String, _
        ByVal tableName as String ) Export as DWord

   Dim i as Long
   Dim iCount as Long
   Dim iReturn as Long
   Dim sAccum as String
   Dim zIndex as Asciiz * 300
   Dim zDesc as Asciiz * 300
   Dim zKeys as Asciiz * 3000
   Dim zParam as Asciiz * 300
   Dim pIndex as LINKED_LIST_TYPE PTR

   '- Select the database
   if dbuse(dbProc, dbName + $nul) = %FAIL then
      function = 0
      exit function
   end if

   '- The name of the stored proc.
   if dbrpcinit(dbProc, "sp_helpindex", 0) = %false then
      function = 0
      exit function
   end if

   '- Specify the parameters
   zParam = trim$(tableName)
   if dbrpcparam(dbProc, "@objname", 0, %SQLVARCHAR, _
         sizeof(zParam), len(zparam), ByVal VarPtr(zParam)) = %FAIL then
      function = 0
      exit function
   end if

   '- Execute it
   if dbrpcexec(dbProc) = %FAIL then
      function = 0
      exit function
   end if

   '- Process the results
   iReturn = dbresults(dbProc)
   if iReturn <> %SUCCEED then
      dbcanquery dbProc
      function = %false
      exit function
   end if

   '- Bind the column that we care about
   For i = 1 to dbnumcols(dbProc)
      select case trim$(lcase$(dblibColname(dbProc, i)))
         Case "index_name"
            dbbind dbProc, i, %NTBSTRINGBIND, 0, zIndex
         Case "index_description"
            dbbind dbProc, i, %NTBSTRINGBIND, 0, zDesc
         Case "index_keys"
            dbbind dbProc, i, %NTBSTRINGBIND, 0, zKeys
      end select
   Next i

   iCount = 0
   iReturn = dbnextrow(dbProc)
   Do until iReturn = %NO_MORE_ROWS
      if left$(ucase$(zIndex), 8) <> "_WA_SYS_" then
         incr iCount

         sAccum = trim$(zIndex) + ";" + trim$(zDesc) + ";" + _
                  trim$(zKeys)

         if pIndex = 0 then
            pIndex = llAdd(0, Format$(iCount), sAccum)
         else
            llUpdateByName pIndex, Format$(iCount), sAccum
         end if
      end if
      iReturn = dbnextrow(dbProc)
   Loop

   Function = pIndex

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibTable2xml
'
'  Returns the structure of a sql server table in pappy:database-structure
'  xml format.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function dblibTable2xml Alias "dblibTable2xml" _
      ( ByVal dbProc as Long, ByVal sDb as String, ByVal sTable as String ) _
        Export as String

   Dim pFields as Dword
   Dim pIndex as Dword
   Dim pNext as Dword
   Dim sField as String
   Dim sIndex as String
   Dim sXML as String

   sXML = "  <" + $DBLIB_TABLE + " name='" + _
          xmlClean(sTable) + "'>" + $CRLF

   pFields = dblibEnumFields(dbProc, sDB, sTable)
   pNext = pFields
   do until pNext = 0
      sField = llData(pNext)
      sXML = sXML + "    <" + $DBLIB_FIELD + " " + _
            "type='" + parse$(sField, ";", 2) + "' " + _
            "length='" + parse$(sField, ";", 3) + "' " + _
            "precision='" + parse$(sField, ";", 4) + "' " + _
            "scale='" + parse$(sField, ";", 5) + "' " + _
            "nulls='" + parse$(sField, ";", 6) + "' " + _
            "column='" + parse$(sField, ";", 1) + "'>" + _
            Parse$(sField, ";", 1) + _
            "</" + $DBLIB_FIELD + ">" + $crlf
      pNext = llNext(pNext)
   loop

   pIndex = dblibEnumIndices(dbProc, sDB, sTable)
   pNext = pIndex
   do until pNext = 0
      sIndex = llData(pNext)
      sXML = sXML + "    <" + $DBLIB_INDEX + " " + _
            "desc='" + parse$(sIndex, ";", 2) + "' " + _
            "fields='" + parse$(sIndex, ";", 3) + "' " + _
            "name='" + parse$(sIndex, ";", 1) + "'>" + _
            parse$(sIndex, ";", 1) + _
            "</" + $DBLIB_INDEX + ">" + $crlf
      pNext = llNext(pNext)
   loop
   sXML = sXML + "  </" + $DBLIB_TABLE + ">" + $crlf

   if pFields then llFree pFields
   if pIndex then llFree pIndex
   Function = sXML

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibDatabase2xml
'
'  Puts a database structure into XML, according to the ...
'  pappy:xml-database-structure format used by WinRep Software
'  to describe their SQL Server and Access databases during
'  database structure replication.
'
'  Returns a the XML structure in string form if successful. If not,
'  the "" is returned and LastError is set.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function dblibDatabase2xml Alias "dblibDatabase2xml" _
      ( ByVal dbProc as Long, ByVal sDB as String ) export as String

   Dim xmlReturn as Dword
   Dim pTables as Dword
   Dim pNext as Dword
   Dim pFields as Dword
   Dim pNextField as Dword
   Dim pIndex as Dword
   Dim pNextIndex as Dword
   Dim sTable as String
   Dim sErr as String
   Dim sXML as String

   sXML = "<" + $DBLIB_DATABASE + " name='" + sDB + "'>" + $crlf

   '- Get a list of tables
   pTables = dblibEnumTables(dbProc, sDB)
   if pTables = 0 then
      sErr = "
      Goto dblib2xml_Error
   end if

   '- Get info for each table
   pNext = pTables
   Do until pNext = 0

      sTable = llData(pNext)
      sXML = sXML + $crlf + dblibTable2XML(dbProc, sDb, sTable)
      pNext = llNext(pNext)
   Loop

   if pTables then llFree pTables

   sXML = sXML + $crlf + "</" + $DBLIB_DATABASE + ">"

   Function = sXML
   exit function

dblib2xml_Error:
   Function = ""

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibBuildInsertStatement
'
'  Takes a linked list of field names/values and turns it into
'  an sql insert statement
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function dblibBuildInsertStatement alias "dblibBuildInsertStatement" _
      ( ByVal sTable as String, listFields as LINKED_LIST_TYPE PTR ) as String

   Dim pNext as Long
   Dim sSQL as String
   Dim sAccum as String
   Dim oneName as String
   Dim oneValue as String
   Dim allNames as String
   Dim allValues as String

   '- Build the insert statement from the fields
   allNames = ""
   allValues = ""
   pNext = listFields
   do until pNext = 0
      oneName = trim$(lcase$(llName(pNext)))
      if len(oneName) then

         oneValue = trim$(llData(pNext))
         if left$(oneName, 3) = ":n:" then
            oneName = mid$(oneName, 4)
            oneValue = format$(val(oneValue))

         elseif (right$(oneName, 2) = "_m") or _
            (right$(oneName, 4) = "_int") or _
            (right$(oneName, 2) = "_r") or _
            (right$(oneName, 2) = "_f") or _
            (right$(oneName, 4) = "_amt") then

            oneValue = format$(val(oneValue))

         elseif oneValue = "" then
            oneValue = "null"

         else
            oneValue = "'" + dblibClean(oneValue) + "'"

         end if

         if allNames = "" then
            allNames = oneName
         else
            allNames = allNames + "," + oneName
         end if

         if allValues = "" then
            allValues = oneValue
         else
            allValues = allValues + "," + oneValue
         end if
      end if
      pNext = llNext(pNext)
   loop

   function = "INSERT INTO " + sTable + _
         " (" + allNames + ") " + _
         "VALUES (" + allValues + ")"

end function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibInsert
'
'  Builds and executes an sql insert statement from the database name,
'  table name, and linked list of fields provided.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function dblibInsert Alias "dblibInsert" _
      ( ByVal dbProc as Long, ByVal sDB as String, ByVal sTable as String, _
        listFields as LINKED_LIST_TYPE PTR ) export as Long

   Dim sSQL as String

   if listFields = 0 then
      dblibSetError dbProc, "dblibInsert: No fields sent"
      function = %false
      exit function
   end if

   sSQL = dblibBuildInsertStatement(sTable, listFields)
   llAdd listFields, "dblibInsert:sql", sSQL

   function = dblibExecuteNoReturn(dbProc, sDB, sSQL)

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  sql2xmlAttrib
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function sql2xmlAttrib alias "sql2xmlAttrib" _
      ( ByVal pReturn as Dword ) export as string

   Dim pNext as dword
   Dim oneName as String
   Dim oneAttrib as String
   Dim sAccum as String

   pNext = llNext(pReturn)
   do until pNext = 0
      oneName = llName(pNext)
      oneAttrib = oneName + "=" + _
            $dq + xmlClean(dblibField(pReturn, oneName)) + $dq
      if sAccum = "" then
         sAccum = oneAttrib
      else
         sAccum = sAccum + " " + oneAttrib
      end if
      pNext = llNext(pNext)
   loop

   function = sAccum

end function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibBeginTrans
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
sub dblibBeginTrans alias "dblibBeginTrans" (ByVal hDB as Long) export

   dbcmd hDB, "BEGIN TRANSACTION;" + $crlf + $nul

end sub

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibAddTrans
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
sub dblibAddTrans alias "dblibAddTrans" _
      ( ByVal hDB as Long, ByVal sSQL as String ) export

   dbcmd hDB, sSQL + ";" + $crlf + $nul

end sub

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibEndTrans
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function dblibCommitTrans alias "dblibCommitTrans" _
      (ByVal hDB as Long) export as Long

   Dim iResults as Long
   Dim iNextRow as Long

   dbcmd hDB, "COMMIT TRANSACTION;" + $nul

   '- Execute the sql statement
   if dbsqlexec(hDB) = %FAIL then
      function = %false
      exit function
   end if

   '- Prep the results
   iResults = dbresults(hDB)
   do until iResults = %NO_MORE_RESULTS
      if iResults = %FAIL then
         dbcancel hDB
         function = %false
         exit function
      end if

      iNextRow = dbnextrow(hDB)
      do until iNextRow = %NO_MORE_ROWS
         if iNextRow = %FAIL then
            dbcancel hDB
            function = %false
            exit function
         end if
         iNextRow = dbnextrow(hDB)
      loop
      iResults = dbresults(hDB)
   loop

   function = %true

end function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibCrackDate
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
sub dblibCrackDate alias "dblibCrackDate" _
      ( sDate as String, d as DBLIB_DATE_TYPE )

   Dim iMod as Long
   Dim tm as String

   sDate = lcase$(sDate)
   Select Case left$(sDate, 3)
      Case "jan": d.m = 1
      Case "feb": d.m = 2
      Case "mar": d.m = 3
      Case "apr": d.m = 4
      Case "may": d.m = 5
      Case "jun": d.m = 6
      Case "jul": d.m = 7
      Case "aug": d.m = 8
      Case "sep": d.m = 9
      Case "oct": d.m = 10
      Case "nov": d.m = 11
      Case "dec": d.m = 12
      Case else
         d.m = 1
   End Select
   d.d = int(val(parse$(sDate, " ", 2)))
   d.y = int(val(parse$(sDate, " ", 3)))
   Replace "  " with " " in sDate
   tm = trim$(parse$(sDate, " ", 4))
   if right$(tm, 2) = "pm" then
      iMod = 12
   elseif right$(tm, 2) = "am" then
      iMod = 0
   else
      iMod = -1
   end if
   d.hr = int(val(extract$(tm, ":")))
   d.mn = int(val(left$(remain$(tm, ":"), 2)))
   d.sc = 0

   if (iMod = 0) or (iMod = 12) then
      if d.hr < 12 then
         d.hr = d.hr + iMod
      elseif (d.hr = 12) and (iMod = 0) then
         d.hr = 0
      end if
   end if
end sub

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  dblibQuickDate
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function dblibQuickDate(ByVal sDate as String) as String

   Dim d as DBLIB_DATE_TYPE

   dblibCrackDate sDate, d
   function = format$(d.m) + "/" + format$(d.d) + "/" + format$(d.y, "0000")

end function

#endif
