'
'  test.bas
'
'  dblib test program for pbcc
'  Compile with pbcc2.1 or greater. will not work correctly
'  with pbcc2.0 or earlier (even if it compiles).
'
#compile exe
#dim all

#include "win32api.inc"
#include "dd_dblib.bas"

'- IMPORTANT
'  You'll need to change these contants to work with your computer
$SQL_SERVER = "192.168.10.1"
$SQL_DB = "wradmin"
$SQL_USER = "sa"
$SQL_PASSWORD = "password"
$SQL_TABLE = "t_don_data"

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  pbmain
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function pbmain

   Dim i as Long
   Dim bFound as Long
   Dim hDB as Long
   Dim pFields as Long
   Dim sSQL as String

   '- Always call this first in your program
   dblibInit

   '- Connect to a server and open a database.
   '  If the database name is blank, then the default is selected.
   '  If the database name is invalid, this should fail.
   '
   hDB = dblibUseDB($SQL_SERVER, $SQL_DB, $SQL_USER, $SQL_PASSWORD)
   if hDB = 0 then
      Stdout "Error: " + dblibLastError(0)
      goto pbmain_done
   end if

   '- Create a table if it isn't there
   if dblibTableExists(hDB, $SQL_DB, $SQL_TABLE) = %false then
      stdout "creating table " + $SQL_TABLE
      sSQL = "CREATE TABLE " + $SQL_TABLE + " (name VARCHAR(50) null, " + _
            "id VARCHAR(10) null);"
      if dblibExecuteNoReturn(hDB, $SQL_DB, sSQL) = %false then
         stdout "Unable to create table " + $SQL_TABLE
         goto pbmain_done
      end if
      stdout "table created"

   end if

   '- Delete any existing records using dblibExecuteNoReturn
   '  This function executes an sql statement that expects
   '  no return value - inserts, updates, deletes. It returns
   '  %true on success, %false on failure.
   '
   sSQL = "DELETE FROM " + $SQL_TABLE + " WHERE 1=1"
   if dblibExecuteNoReturn(hDB, $SQL_DB, sSQL) = %false then
      stdout "Error deleting existing records from " + $SQL_TABLE
      goto pbmain_done
   end if

   '- Insert some records inside of a transaction
   dblibBeginTrans hDB
   for i = 1 to 20
      sSQL = "INSERT INTO " + $SQL_TABLE + " (name, id) " + _
            "VALUES ('Name " + format$(i) + "', '" + format$(i, "00") + "')"
      dblibAddTrans hDB, sSQL
   next i

   '- All of the added transactions either fail or succeed in bulk, atomicly.
   if dblibCommitTrans(hDB) = %false then
      stdout "Error inserting records. Transaction failed, no records inserted."
      goto pbmain_done
   end if

   '- Find a single record
   sSQL = "SELECT * FROM " + $SQL_TABLE + " WHERE id = '02'"
   pFields = dblibExecuteOne(hDB, $SQL_DB, sSQL, bFound)
   if pFields = 0 then
      stdout "Unable to execute select query."
      stdout "SQL=" + sSQL
      stdout "ERR=" + dblibLastError(hDB)
      goto pbmain_done
   end if

   if bFound = %false then
      stdout "Unable to locate record with ID = '02'"

      '- Free the memory used to return data linked list
      llFree pFields
      goto pbmain_done
   end if

   Stdout "The name of the record with ID = '" + _
      dblibField(pFields, "id") + "' is '" + _
      dblibField(pFields, "name") + "'"

   '- Free returned linked list
   llFree pFields

   '- Select some records and loop through them
   sSQL = "SELECT * FROM " + $SQL_TABLE
   pFields = dblibExecute(hDB, $SQL_DB, sSQL)
   if pFields = 0 then
      stdout "Unable to execute select query."
      stdout "SQL=" + sSQL
      stdout "ERR=" + dblibLastError(hDB)
      goto pbmain_done
   end if

   Stdout "Below are the rows"
   do until dbnextrow(hDB) = %NO_MORE_ROWS
      Stdout dblibField(pFields, "ID") + ": " + dblibField(pFields, "name")
   loop
   llFree pFields

   Stdout "SUCCESS!"

pbmain_done:

   '- This is a direct call to dblib that closes the open database.
   if hDB then
      dbclose hDB
   end if

   '- Always call this last in your program to free memory
   '  and the critical section used for thread safety.
   '
   dblibDone

   waitkey$

end function
